/** <title>NSSound</title>

   <abstract>Load, manipulate and play sounds</abstract>

   Copyright (C) 2002 Free Software Foundation, Inc.
   
   Author: Enrico Sersale <enrico@imago.ro>
   Date: Jul 2002

   This file is part of the GNUstep GUI Library.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; see the file COPYING.LIB.
   If not, see <http://www.gnu.org/licenses/> or write to the 
   Free Software Foundation, 51 Franklin Street, Fifth Floor, 
   Boston, MA 02110-1301, USA.
*/ 

#include <Foundation/Foundation.h>
#include <unistd.h>

#include "NSSound_test.h"
#include "GSSoundSource.h"
#include "GSSoundSink.h"

/* Private NSConditionLock conditions used for streaming */
enum
{
  SOUND_SHOULD_PLAY = 1,
  SOUND_SHOULD_PAUSE
};

#define BUFFER_SIZE 4096

/* Class variables and functions for class methods */
static NSMutableDictionary *nameDict = nil;
static NSDictionary *nsmapping = nil;
static NSArray *sourcePlugIns = nil;
static NSArray *sinkPlugIns = nil;

static inline void _loadNSSoundPlugIns (void)
{
  NSString      *path;
  NSArray       *paths;
  NSBundle      *bundle;
  NSEnumerator  *enumerator;
  NSMutableArray *all,
                 *_sourcePlugIns,
                 *_sinkPlugIns;
  Class plugInClass;

  /* Gather up the paths */
  paths = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory,
            NSAllDomainsMask, YES);

  enumerator = [paths objectEnumerator];
  all = [NSMutableArray arrayWithCapacity: 10];
  while ((path = [enumerator nextObject]) != nil)
   {
     bundle = [NSBundle bundleWithPath: path];
     paths = [bundle pathsForResourcesOfType: @"nssound"
                                 inDirectory: @"Bundles"];
     [all addObjectsFromArray: paths];
   }
  
  enumerator = [all objectEnumerator];
  _sourcePlugIns = [NSMutableArray arrayWithCapacity: 5];
  _sinkPlugIns = [NSMutableArray arrayWithCapacity: 5];
  while ((path = [enumerator nextObject]) != nil)
    {
      NSBundle *nssoundBundle = [NSBundle bundleWithPath: path];
      plugInClass = [nssoundBundle principalClass];
      if ([plugInClass conformsToProtocol: @protocol(GSSoundSource)])
        {
          [_sourcePlugIns addObject:plugInClass];
        }
      else if ([plugInClass conformsToProtocol: @protocol(GSSoundSink)])
        {
          [_sinkPlugIns addObject:plugInClass];
        }
      else
        {
          NSLog (@"Bundle %@ does not conform to GSSoundSource or GSSoundSink",
            path);
        }
    }
  
  sourcePlugIns = [NSArray arrayWithArray: _sourcePlugIns];
  sinkPlugIns = [NSArray arrayWithArray: _sinkPlugIns];
}



@implementation NSBundle (NSSoundAdditions)

- (NSString *) pathForSoundResource: (NSString *)name
{
  NSString *ext = [name pathExtension];
  NSString *path = nil;

  if ((ext == nil) || [ext isEqualToString:@""])
    {
      NSArray	*types = [NSSound soundUnfilteredFileTypes];
      unsigned	c = [types count];
      unsigned	i;

      for (i = 0; path == nil && i < c; i++)
	      {
	        ext = [types objectAtIndex: i];
	        path = [self pathForResource: name ofType: ext];
	      }
    }
  else
    {
      name = [name stringByDeletingPathExtension];
      path = [self pathForResource: name ofType: ext];
    }
  return path;
}

@end 

@interface NSSound (PrivateMethods)

- (void)_stream;
- (void)_finished: (NSNumber *)finishedPlaying;

@end

@implementation NSSound (PrivateMethods)

- (void)_stream
{
  NSUInteger bytesRead;
  BOOL success = YES;
  void *buffer;
  
  id<GSSoundSource> source = (id<GSSoundSource>)_private[0];
  id<GSSoundSink> sink = (id<GSSoundSink>)_private[1];
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // Open sink for writing.
  [sink open];
  
  // Allocate space for buffer and start writing.
  buffer = NSZoneMalloc(NSDefaultMallocZone(), BUFFER_SIZE);
  bytesRead = [source readBytes: buffer length: BUFFER_SIZE];
  do
    {
      while ((!_shouldStop) && (bytesRead > 0) && success)
        {
          [lock lockWhenCondition: SOUND_SHOULD_PLAY];
          [lock unlock];
          if (_shouldStop)
            {
              break;
            }
          success = [sink writeBytes: buffer length: bytesRead];
          bytesRead = [source readBytes: buffer length: BUFFER_SIZE];
        }
      
      [source setCurrentTime: 0.0];
    } while (_loops == YES && !_shouldStop);
  
  /* FIXME: -performSelectorOnMainThread:withObject:waitUntilDone: is not
       calling -_finished: when called from here */
  [self performSelector: @selector(_finished:)
             withObject: [NSNumber numberWithBool: success]];
  
  NSZoneFree (NSDefaultMallocZone(), buffer);
  [sink close];
}

- (void)_finished: (NSNumber *)finishedPlaying
{
  NSConditionLock *lock = (NSConditionLock *)_private[2];
  DESTROY(lock);
  _private[2] = (void *)lock;
  
  if (_delegate)
    {
      [_delegate sound: self didFinishPlaying: [finishedPlaying boolValue]];
    }
}

@end

@implementation	NSSound

+ (void) initialize
{
  if (self == [NSSound class])
    {
      NSString *path = [NSBundle pathForLibraryResource: @"nsmapping"
	                        			                 ofType: @"strings"
				                                    inDirectory: @"Sounds"];
      [self setVersion: 2];

      nameDict = [[NSMutableDictionary alloc] initWithCapacity: 10];
      
      if (path)
        {
          nsmapping = RETAIN([[NSString stringWithContentsOfFile: path]
                  propertyListFromStringsFileFormat]);
        }
      
      _loadNSSoundPlugIns ();
    }
}

- (void) dealloc
{
  // Make sure sound is stopped before deallocating.
  [self stop];
  
  TEST_RELEASE (_data);
  if ((_name != nil) && self == [nameDict objectForKey: _name])
    { 
      [nameDict removeObjectForKey: _name];
    }
  TEST_RELEASE (_name);
  TEST_RELEASE (_playbackDeviceIdentifier);
  TEST_RELEASE (_channelMapping);
  TEST_RELEASE ((id<GSSoundSource>)_private[0]);
  TEST_RELEASE ((id<GSSoundSink>)_private[1]);
  
  [super dealloc];
}

//
// Creating an NSSound 
//
- (id) initWithContentsOfFile: (NSString *)path byReference:(BOOL)byRef
{
  NSData *fileData;
  
  _onlyReference = byRef;
  ASSIGN (_name, [path lastPathComponent]);
  [nameDict setObject: self forKey: _name];
  
  fileData = [NSData dataWithContentsOfMappedFile: path];
  if (!fileData)
    {
      NSLog (@"Could not get sound data from: %@", path);
      DESTROY(self);
      return nil;
    }
  
  return [self initWithData: fileData];
}

- (id) initWithContentsOfURL: (NSURL *)url byReference:(BOOL)byRef
{
  _onlyReference = byRef;	
  return [self initWithData: [NSData dataWithContentsOfURL: url]];
}

- (id) initWithData: (NSData *)data;
{
  id<GSSoundSource> source;
  id<GSSoundSink> sink;
  NSEnumerator *enumerator;
  Class sourceClass,
        sinkClass;
    
  _data = data;
  RETAIN(_data);
  
  enumerator = [sourcePlugIns objectEnumerator];
  NSLog (@"%@", enumerator);
  while ((sourceClass = [enumerator nextObject]) != nil)
    {
      NSLog (@"%@", sourceClass);
      if ([sourceClass canInitWithData: _data])
        {
          source = [[sourceClass alloc] initWithData: _data];
          if (source == nil)
            {
              NSLog (@"Could not read sound data!");
              DESTROY(self);
              return nil;
            }
          _private[0] = (void *)source;
          
          break;
        }
    }
  
  enumerator = [sinkPlugIns objectEnumerator];
  NSLog (@"%@", enumerator);
  while ((sinkClass = [enumerator nextObject]) != nil)
    {
      NSLog (@"%@", sinkClass);
      if ([sinkClass canInitWithPlaybackDevice: nil])
        {
          sink = [[sinkClass alloc] initWithEncoding: [source encoding]
                                            channels: [source channelCount]
                                          sampleRate: [source sampleRate]
                                           byteOrder: [source byteOrder]];
          if (sink == nil)
            {
              NSLog (@"Could not open sound sink!");
              DESTROY(self);
              return nil;
            }
          _private[1] = (void *)sink;
          
          break;
        }
    }
  
  if (sourceClass == nil || sinkClass == nil)
    {
      NSLog (@"Could not find suitable sound plug-in");
      DESTROY(self);
      return nil;
    }
  
  return self;
}

//
// Playing and Information
//
- (BOOL) pause 
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // Do nothing if sound is already paused.
  if ([lock condition] == SOUND_SHOULD_PAUSE)
    {
      return NO;
    }
  
//  [lock lock];
  if ([lock tryLock] == NO)
    {
      return NO;
    }
  [lock unlockWithCondition: SOUND_SHOULD_PAUSE];
  return YES;
}

- (BOOL) play
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // If the NSCondtionLock exists it's because NSSound is already playing
  if (lock != nil)
    {
      return NO;
    }
  
  lock = [[NSConditionLock alloc] initWithCondition: SOUND_SHOULD_PAUSE];
  _private[2] = (void *)lock;
  if ([lock tryLock] != YES)
    {
      return NO;
    }
  // Set _shouldStop = NO before starting.
  _shouldStop = NO;
  [NSThread detachNewThreadSelector: @selector(_stream)
                           toTarget: self
                         withObject: nil];
  [lock unlockWithCondition: SOUND_SHOULD_PLAY];
  
  return YES;
}

- (BOOL) resume
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // Do nothing if sound is already playing.
  if ([lock condition] == SOUND_SHOULD_PLAY)
    {
      return NO;
    }
  
  if ([lock tryLock] == NO)
    {
      return NO;
    }
  [lock unlockWithCondition: SOUND_SHOULD_PLAY];
  return YES;
}

- (BOOL) stop
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  if (lock == nil)
    {
      return NO;
    }
  
  if ([lock tryLock] != YES)
    {
      return NO;
    }
  _shouldStop = YES;
  // Set to SOUND_SHOULD_PLAY so that thread isn't blocked.
  [lock unlockWithCondition: SOUND_SHOULD_PLAY];
  
  return YES;
}

- (BOOL) isPlaying
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  if (lock == nil)
    {
      return NO;
    }
  if ([lock condition] == SOUND_SHOULD_PLAY)
    {
      return YES;
    }
  return NO;
}

- (float) volume
{
  return [(id<GSSoundSource>)_private[0] volume];
}

- (void) setVolume: (float) volume
{
  [(id<GSSoundSource>)_private[0] setVolume: volume];
}

- (NSTimeInterval) currentTime;
{
  return [(id<GSSoundSource>)_private[0] currentTime];
}

- (void) setCurrentTime: (NSTimeInterval) currentTime
{
  /* FIXME: this current crashes apps */
  /* NSConditionLock *lock = (NSConditionLock *)_private[2];
  [lock lock];
  [(id<GSSoundSource>)_private[0] setCurrentTime: currentTime];
  [lock unlock]; */
}

- (BOOL) loops;
{
  return NO;
}

- (void) setLoops: (BOOL) loops
{
  return;
}

- (NSTimeInterval) duration
{
  id<GSSoundSource> source = (id<GSSoundSource>)_private[0];
  return [source duration];
}

//
// Working with delegates 
//
- (id) delegate
{
  return _delegate;
}

- (void) setDelegate: (id)aDelegate
{
  _delegate = aDelegate;
}

//
// Naming Sounds 
//
+ (id) soundNamed: (NSString*)name
{
  NSString	*realName = [nsmapping objectForKey: name];
  NSSound	*sound;

  if (realName)
    {
      name = realName;
    }
	
  sound = (NSSound *)[nameDict objectForKey: name];
 
  if (sound == nil)
    {
      NSString	*extension;
      NSString	*path = nil;
      NSBundle	*main_bundle;
      NSArray	*array;
      NSString	*the_name = name;

      // FIXME: This should use [NSBundle pathForSoundResource], but this will 
      // only allow soundUnfilteredFileTypes.
      /* If there is no sound with that name, search in the main bundle */
			
      main_bundle = [NSBundle mainBundle];
      extension = [name pathExtension];
		
      if (extension != nil && [extension length] == 0)
	{
	  extension = nil;
	}

      /* Check if extension is one of the sound types */
      array = [NSSound soundUnfilteredFileTypes];
	
      if ([array indexOfObject: extension] != NSNotFound)
	{
	  /* Extension is one of the sound types
	     So remove from the name */
	  the_name = [name stringByDeletingPathExtension];
	} 
      else 
	{
	  /* Otherwise extension is not an sound type
	     So leave it alone */
	  the_name = name;
	  extension = nil;
	}

      /* First search locally */
      if (extension)
	{
	  path = [main_bundle pathForResource: the_name ofType: extension];
	} 
      else 
	{
	  id o, e;

	  e = [array objectEnumerator];
	  while ((o = [e nextObject]))
	    {
	      path = [main_bundle pathForResource: the_name ofType: o];
	      if (path != nil && [path length] != 0)
		{
		  break;
		}
	    }
	}

      /* If not found then search in system */
      if (!path)
	{
	  if (extension)
	    {
	      path = [NSBundle pathForLibraryResource: the_name
				               ofType: extension
				          inDirectory: @"Sounds"];
	    } 
	  else 
	    {
	      id o, e;

	      e = [array objectEnumerator];
	      while ((o = [e nextObject])) {
	      path = [NSBundle pathForLibraryResource: the_name
				               ofType: o
				          inDirectory: @"Sounds"];
		if (path != nil && [path length] != 0)
		  {
		    break;
		  }
	      }
	    }
	}

      if ([path length] != 0)
	{
	  sound = [[self allocWithZone: NSDefaultMallocZone()]
		    initWithContentsOfFile: path byReference: NO];

	  if (sound != nil)
	    {
	      [sound setName: name];
	      RELEASE(sound);	
	      sound->_onlyReference = YES;
	    }

	  return sound;
	}
    }  
	
  return sound;
}

+ (NSArray *) soundUnfilteredFileTypes
{
  Class sourceClass;
  NSMutableArray *array = [NSMutableArray arrayWithCapacity: 10];
  NSEnumerator *enumerator = [sourcePlugIns objectEnumerator];
  while ((sourceClass = [enumerator nextObject]) != nil)
    {
      NSArray *output = [sourceClass soundUnfilteredFileTypes];
      [array addObjectsFromArray: output];
    }
  
  return array;
}

+ (NSArray *) soundUnfilteredTypes
{
  Class sourceClass;
  NSMutableArray *array = [NSMutableArray arrayWithCapacity: 10];;
  NSEnumerator *enumerator = [sourcePlugIns objectEnumerator];
  while ((sourceClass = [enumerator nextObject]) != nil)
    {
      [array addObjectsFromArray: [sourceClass soundUnfilteredTypes]];
    }
  
  return array;
}

- (NSString *) name
{
  return _name;
}

- (BOOL) setName: (NSString *)aName
{
  BOOL retained = NO;
  
  if (!aName || [nameDict objectForKey: aName])
    {
      return NO;
    }
	
  if ((_name != nil) && self == [nameDict objectForKey: _name])
    {
      /* We retain self in case removing from the dictionary releases us */
      RETAIN (self);
      retained = YES;
      [nameDict removeObjectForKey: _name];
    }
  
  ASSIGN(_name, aName);
  
  [nameDict setObject: self forKey: _name];
  if (retained)
    {
      RELEASE (self);
    }
  
  return YES;
}

- (NSString *) playbackDeviceIdentifier
{
  return [(id<GSSoundSink>)_private[1] playbackDeviceIdentifier];
}

- (void) setPlaybackDeviceIdentifier: (NSString *)playbackDeviceIdentifier
{
  id<GSSoundSource> source = (id<GSSoundSource>)_private[0];
  id<GSSoundSink> sink = (id<GSSoundSink>)_private[1];
  NSConditionLock *lock = (NSConditionLock *)_private[2];
  
  if ([[sink class] canInitWithPlaybackDevice: playbackDeviceIdentifier])
    {
      [sink setPlaybackDeviceIdentifier: playbackDeviceIdentifier];
    }
  else
    {
      Class sinkClass;
      NSEnumerator *enumerator = [sinkPlugIns objectEnumerator];
      while ((sinkClass = [enumerator nextObject]) != nil)
        {
          if ([sinkClass canInitWithPlaybackDevice: playbackDeviceIdentifier])
            {
              [lock lock];
              [sink close];
              DESTROY(sink);
              sink = [[sinkClass alloc] initWithEncoding: [source encoding]
                                                channels: [source channelCount]
                                              sampleRate: [source sampleRate]
                                               byteOrder: [source byteOrder]];
              [sink open];
              _private[1] = (void *)sink;
              [lock unlock];
            }
        }
    }
}

- (NSArray *) channelMapping
{
  return [(id<GSSoundSink>)_private[1] channelMapping];
}

- (void) setChannelMapping: (NSArray *)channelMapping
{
  [(id<GSSoundSink>)_private[1] setChannelMapping: channelMapping];
}

@end
